///Copyright 2018 Miolin

///Licensed under the Apache License, Version 2.0 (the "License");
///you may not use this file except in compliance with the License.
///You may obtain a copy of the License at

///http://www.apache.org/licenses/LICENSE-2.0

///Unless required by applicable law or agreed to in writing, software
///distributed under the License is distributed on an "AS IS" BASIS,
///WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
///See the License for the specific language governing permissions and
///limitations under the License.

library state_controller;

import 'package:flutter/widgets.dart';
import 'dart:collection';

abstract class StateWithController<W extends StatefulWidget,
    T extends StateController> = State<W> with ControllerInjector<W, T>;

mixin ControllerInjector<W extends StatefulWidget, T extends StateController>
    on State<W> {
  T _controller;

  T get controller => _controller;

  @protected
  @mustCallSuper
  void initState() {
    _controller = ControllerProvider.instance.provide<T>();
    _controller._context = context;
    _controller.init();
    configureController(_controller);
    super.initState();
  }

  @protected
  @mustCallSuper
  void didUpdateWidget(covariant W oldWidget) {
    _controller._context = context;
    configureController(_controller);
    _controller.onConfigurationChanged();
    super.didUpdateWidget(oldWidget);
  }

  @protected
  void configureController(T controller) {}

  @protected
  @mustCallSuper
  void dispose() {
    super.dispose();
    _controller.dispose();
    _controller = null;
  }
}

class ControllerProviderConfig extends InheritedWidget {
  final List<Configuration> configurations;
  final Widget child;

  ControllerProviderConfig(
      {@required this.configurations, @required this.child, Key key})
      : super(key: key, child: child) {
    if (ControllerProvider.instance._configurations.isEmpty) {
      _updateConfig(configurations);
    }
  }

  static void _updateConfig(
      List<Configuration<StateController>> configurations) {
    configurations.forEach((conf) =>
        ControllerProvider.instance.addControllerConfiguration(conf));
  }

  @override
  bool updateShouldNotify(ControllerProviderConfig oldWidget) {
    var isUpdate = oldWidget.configurations != configurations;
    _updateConfig(configurations);
    return isUpdate;
  }
}

class ControllerProvider {
  final HashMap<Type, Configuration> _configurations = HashMap();
  static final ControllerProvider _provider = ControllerProvider._();

  static ControllerProvider get instance => _provider;

  ControllerProvider._();

  void addControllerConfiguration(Configuration configuration) {
    if (!_configurations.containsKey(configuration.type)) {
      _configurations.addAll({configuration.type: configuration});
    } else {
      print(
          'You cannot add a type ${configuration.type}, it is already in the configuration.');
    }
  }

  T provide<T extends StateController>() {
    assert(T != dynamic);
    assert(_configurations.containsKey(T));
    return _configurations[T].provide();
  }
}

typedef ControllerCreator<T extends StateController>();

class Configuration<T extends StateController> {
  final singleton;
  final initOnStart;
  final Type type;
  final ControllerCreator<T> creator;
  T _instance;

  Configuration({
    @required this.type,
    @required this.creator,
    this.singleton = true,
    this.initOnStart = false,
  }) {
    if (initOnStart && singleton) {
      _instance = creator();
    }
  }

  T provide() {
    if (singleton) {
      if (_instance == null)
        _instance = creator();
      return _instance;
    } else {
      return creator();
    }
  }
}

abstract class StateController {
  BuildContext _context;

  BuildContext get context => _context;

  @protected
  void init() {}

  @protected
  void onConfigurationChanged() {}

  @protected
  @mustCallSuper
  void dispose() {
    _context = null;
  }
}
